﻿B4J=true
Group=Default Group
ModulesStructureVersion=1
Type=Class
Version=6.3
@EndOfDesignText@
'version 0.9
Sub Class_Globals
	Private xui As XUI
	Public gs As X2GameStep
	Type X2GameStep (BodiesToDelete As List, GameTimeMs As Int, _
		 ShouldDraw As Boolean, _
		DrawingTasks As List)
	Private Index As Int
	Public TargetView As B4XView 'Foreground ImageView
	Public MainBC As BitmapCreator  'Foreground BC
	'Used to schedule a future task. See X2.AddFutureTask.
	Type X2FutureTask (Callback As Object, SubName As String, GameTimeMs As Int, Value As Object)
	Private FutureTasks As List
	Private LoopsPerSecond As Float
	Public IsRunning As Boolean
	Public mGame As Game
	Public World As B2World
	Public mBCPixelsPerMeter As Float
	Public GraphicCache As X2SpriteGraphicCache
	Public BmpSmoothScale As Float = 1
	'Visible window
	Public ScreenAABB As B2AABB
	'A bitmap and its scale.
	Type X2ScaledBitmap (Bmp As B4XBitmap, Scale As Float)
	Private Drawing As Boolean
	#if debug
	Public TargetFPS As Int = 30
	#Else
	Public TargetFPS As Int = 60
	#end if
	Private Transparent As BitmapCreator
	'The duration of each time step.
	Public TimeStepMs As Float 
	Private DebugDraw As X2DebugDraw
	Private IsDebugDrawEnabled As Boolean
	Public SoundPool As X2SoundPool
	'True when drawings are throtlled.
	Public IsLowFPS As Boolean
	Public FPS As Float
	Private LastDrawingTime As Long
	'Decreases or increases the game speed. Call UpdateTimeParameters to apply changes.
	Public SlowDownPhysicsScale As Float = 1
	Private SleepTime As Int
	
End Sub

Public Sub Initialize (MyGame As Game, vTargetView As B4XView)
	mGame = MyGame
	TargetView = vTargetView
	MainBC.Initialize(TargetView.Width / xui.Scale, TargetView.Height / xui.Scale)
	Transparent.Initialize(MainBC.mWidth, MainBC.mHeight)
	ScreenAABB.Initialize
	GraphicCache.Initialize (Me)
	If xui.IsB4J Then
		BmpSmoothScale = 2
	Else
		BmpSmoothScale = xui.Scale
	End If
	World.Initialize("world", World.CreateVec2(0, -10))
	SoundPool.Initialize
	Reset
End Sub

Private Sub MainLoop
	Index = Index + 1
	Dim MyIndex As Int = Index
	Dim SkippedFrames As Int
	LastDrawingTime = DateTime.Now
	FPS = TargetFPS
	Dim counter As Int
	Do While MyIndex = Index
		Dim StartLoopTime As Long = DateTime.Now 'ignore
		World.TimeStep(TimeStepMs / 1000, 2, 3)
		Dim tasks As List
		tasks.Initialize
		Dim dt As DrawTask = MainBC.CreateDrawTask(Transparent, Transparent.TargetRect, 0, 0, True)
		dt.TargetBC = MainBC
		tasks.Add(dt)
		gs.DrawingTasks = tasks
		gs.BodiesToDelete.Clear
		gs.GameTimeMs = gs.GameTimeMs + TimeStepMs
		gs.ShouldDraw = Drawing = False
		Dim NonDrawingIteration As Boolean
		If LoopsPerSecond < TargetFPS - 5 Then
			If IsLowFPS = False Then
				LoopsPerSecond = -1000000
				IsLowFPS = True
			End If
			NonDrawingIteration = counter Mod 2 = 0
		Else
			IsLowFPS = False
		End If
		If NonDrawingIteration Then gs.ShouldDraw = False
		mGame.Tick(gs)
		Dim VisibleBodies As Map = World.QueryAABBToMapOfBodies(ScreenAABB)
		Dim AllBodies As List = World.AllBodies
		For Each body As B2Body In AllBodies
			If body.Tag Is X2BodyWrapper Then
				Dim bw As X2BodyWrapper = body.Tag
				bw.IsVisible = VisibleBodies.ContainsKey(body)
				bw.Tick(gs)
			End If
		Next
		RemoveDeletedBodies
		RunFutureTasks
		If gs.ShouldDraw Then
			Drawing = True
			If IsDebugDrawEnabled Then
				DebugDraw.Draw(gs, VisibleBodies)
			End If
			MainBC.DrawBitmapCreatorsAsync(Me, "BC", gs.DrawingTasks)
		Else If NonDrawingIteration = False Then
			Log("skipping frame!!!")
			SkippedFrames = SkippedFrames + 1
			LoopsPerSecond = LoopsPerSecond - 1
		End If
		
		counter = counter + 1
		If mGame.lblStats.IsInitialized And mGame.lblStats.Visible Then
			Dim stats As String = $"FPS: ${NumberFormat(FPS, 0, 0)}, Time: ${ConvertMillisecondsToString(gs.GameTimeMs)}"$
			stats = stats & $", Bodies: ${AllBodies.Size}, ScreenAABB: ($1.1{ScreenAABB.BottomLeft.X},$1.1{ScreenAABB.BottomLeft.Y})-($1.1{ScreenAABB.TopRight.X},$1.1{ScreenAABB.TopRight.Y})"$
			If IsDebugDrawEnabled Then
				stats = stats & ", DebugDraw!"
			End If
		#if debug
			stats = "DEBUG MODE! " & stats
		#End If
			mGame.lblStats.Text = stats
		End If
		'Log(stats & ", Iteration: " & $"$1.0{IterationTime}"$)
		Sleep(Max(SleepTime - (DateTime.Now - StartLoopTime), 7))
		LoopsPerSecond = (LoopsPerSecond * 20 + 1000/(DateTime.Now - StartLoopTime)) / 21
	Loop
End Sub


Private Sub BC_BitmapReady (bmp As B4XBitmap)
	Drawing = False
	FPS = Min((FPS * 20 + 1000 / (DateTime.Now - LastDrawingTime)) / 21, TargetFPS)
	LastDrawingTime = DateTime.Now
	SetBitmapWithFitOrFill(TargetView, bmp)
	mGame.DrawingComplete
End Sub

'Enabled debug drawing.
Public Sub EnableDebugDraw
	If IsDebugDrawEnabled Then Return
	DebugDraw.Initialize(TargetView.Parent, Me)
	IsDebugDrawEnabled = True
End Sub

'Resets the game state.
Public Sub Reset
	FutureTasks.Initialize
	MainBC.FillRect(xui.Color_Transparent, MainBC.TargetRect)
	gs.Initialize
	gs.GameTimeMs = 0
	gs.BodiesToDelete.Initialize
	UpdateTimeParameters
End Sub

'Call this method after you change TargetFPS or SlowDownPhysicsScale (when the game is running).
Public Sub UpdateTimeParameters
	TimeStepMs = 1000 / TargetFPS
	Dim SleepTime As Int = TimeStepMs
	If xui.IsB4A Then SleepTime = SleepTime - 2
	If xui.IsB4J Then SleepTime = SleepTime - 1
	TimeStepMs = TimeStepMs / SlowDownPhysicsScale
	LoopsPerSecond = TargetFPS
End Sub

'Starts the main loop.
Public Sub Start
	If IsRunning Then Return
	IsRunning = True
	MainLoop
End Sub
'Stops the main loop.
Public Sub Stop
	Index = Index + 1
	IsRunning = False
End Sub

'Creates a B2Body and a X2BodyWrapper.
'Delegate can be Null.
Public Sub CreateBodyAndWrapper (bd As B2BodyDef, Delegate As Object, Name As String) As X2BodyWrapper
	Dim wrapper As X2BodyWrapper
	wrapper.Initialize(mGame, Delegate, Name)
	wrapper.SetBody(World.CreateBody(bd))
	Return wrapper
End Sub

Public Sub DegreesToB2Angle (Degrees As Int) As Float
	Return -cPI / 180 * Degrees
End Sub

Public Sub B2AngleToDegrees (Angle As Float) As Int
	Return -Angle * 180 / cPI
End Sub

'Loads an image file. Resizes it based on the provided dimensions.
'The bitmap scale will be BmpSmoothScale. 
Public Sub LoadBmp (Folder As String, FileName As String, WidthMeters As Float, HeightMeters As Float) As X2ScaledBitmap
	Return LoadBmp2 (Folder, FileName, WidthMeters, HeightMeters, BmpSmoothScale)
End Sub

'Similar to LoadBmp. Allows setting a different scale. Set the scale to 1 if antialiasing is not required.
Public Sub LoadBmp2 (Folder As String, FileName As String, WidthMeters As Float, HeightMeters As Float, BmpScale As Float) As X2ScaledBitmap
	Dim Scale As Float = mBCPixelsPerMeter * BmpScale
	Dim bmp As B4XBitmap = xui.LoadBitmapResize(Folder, FileName, WidthMeters * Scale, HeightMeters * Scale, False)
	Dim sb As X2ScaledBitmap
	sb.Bmp = bmp
	sb.Scale = BmpScale
	Return sb
End Sub

'Splits a sprite sheet bitmap. Returns a list with X2ScaledBitmaps
Public Sub ReadSprites (Bmp As B4XBitmap, Rows As Int, Columns As Int, WidthMeters As Float, HeightMeters As Float) As List
	Dim res As List
	res.Initialize
	Dim scale As Float = mBCPixelsPerMeter * BmpSmoothScale
	Dim RowHeight As Int = Bmp.Height / Rows
	Dim ColumnWidth As Int = Bmp.Width / Columns
	For r = 0 To Rows - 1
		For c = 0 To Columns - 1
			Dim b As B4XBitmap = Bmp.Crop(ColumnWidth * c, RowHeight * r, ColumnWidth, RowHeight).Resize(WidthMeters * scale, HeightMeters * scale, False)
			Dim sb As X2ScaledBitmap
			sb.Bmp = b
			sb.Scale = BmpSmoothScale
			res.Add(sb)
		Next
	Next
	Return res
End Sub

'Configures the coordinates of the visible window.
Public Sub ConfigureDimensions (CenterPointInMeters As B2Vec2, TargetViewWidthInMeters As Float)
	mBCPixelsPerMeter = MainBC.mWidth / TargetViewWidthInMeters
	UpdateWorldCenter(CenterPointInMeters)
End Sub

'Sets the visible window center.
Public Sub UpdateWorldCenter (CenterPointInMeters As B2Vec2)
	ScreenAABB.BottomLeft.X = CenterPointInMeters.X - MainBC.mWidth / 2 / mBCPixelsPerMeter
	ScreenAABB.BottomLeft.Y = CenterPointInMeters.Y - MainBC.mHeight / 2 / mBCPixelsPerMeter
	ScreenAABB.TopRight.X = CenterPointInMeters.X + MainBC.mWidth / 2 / mBCPixelsPerMeter
	ScreenAABB.TopRight.Y = CenterPointInMeters.Y + MainBC.mHeight / 2 / mBCPixelsPerMeter
End Sub

'Converts meters to "pixels". This is not really pixels as the MainBC can also be scaled.
Public Sub MetersToBCPixels (Meters As Float) As Int
	Return Meters * mBCPixelsPerMeter
End Sub

'Converts BC "pixels" to meters.
Public Sub BCPixelsToMeters (Pixels As Int) As Float
	Return Pixels / mBCPixelsPerMeter
End Sub

Private Sub RemoveDeletedBodies
	For Each body As B2Body In gs.BodiesToDelete
		If body.Tag Is X2BodyWrapper Then
			Dim bw As X2BodyWrapper = body.Tag
			bw.Body.Tag = Null
			bw.Body = Null
		End If
		World.DestroyBody(body)
	Next
End Sub

'Converts a world point to a BC pixel.
Public Sub WorldPointToMainBC (x As Float, y As Float) As B2Vec2
	Dim position As B2Vec2
	position.X = (x  - ScreenAABB.BottomLeft.X) * mBCPixelsPerMeter
	position.Y = MainBC.mHeight - 1 - (y - ScreenAABB.BottomLeft.Y) * mBCPixelsPerMeter
	Return position
End Sub
'Converts a screen point to world point. This is useful to convert touch points to the world points.
Public Sub ScreenPointToWorld (x As Int, y As Int) As B2Vec2
	Dim scale As Float = TargetView.Width / MainBC.mWidth * mBCPixelsPerMeter
	Dim position As B2Vec2
	position.X = x / scale + ScreenAABB.BottomLeft.X
	position.Y = (TargetView.Height - 1 - y) / scale + ScreenAABB.BottomLeft.Y
	Return position
End Sub

'Updates the bitmap and sets the scaling mode to FIT in B4J and B4i and FILL in B4A.
'This is useful when the image ratio is the same as the target view but the size is different.
Public Sub SetBitmapWithFitOrFill (vTargetView As B4XView, bmp As B4XBitmap)
	vTargetView.SetBitmap(bmp)
	#if B4A
	'B4XView.SetBitmap sets the gravity in B4A to CENTER. This will prevent the bitmap from being scaled as needed so
	'we switch to FILL
	Dim iv As ImageView = vTargetView
	iv.Gravity = Gravity.FILL
	#End If
End Sub

'Call when the target view is resized. There is assumption that the ImageView ratio is constant.
Public Sub ImageViewResized
	If IsDebugDrawEnabled Then
		DebugDraw.Resize
	End If
End Sub

Private Sub RunFutureTasks
	For i = FutureTasks.Size - 1 To 0 Step - 1
		Dim ft As X2FutureTask = FutureTasks.Get(i)
		If gs.GameTimeMs >= ft.GameTimeMs Then
			CallSub2(ft.Callback, ft.SubName, ft)
			FutureTasks.RemoveAt(i)
		Else
			Exit
		End If
	Next
End Sub


'SubName - Should include an underscore if app is compiled with obfuscation.
'TimeToFire - Don't add gs.GameTime as it will be added automatically.
'Example:<code>
'bw.X2.AddFutureTask(Me, "Add_Score",TimeToLive, score)
'...
'Private Sub Add_Score (ft As X2FutureTask)
'	bw.mGame.mScore.IncreaseScore(ft.Value)
'End Sub</code>
Public Sub AddFutureTask (Callback As Object, SubName As String, TimeToFireMs As Int, Value As Object)
	Dim ft As X2FutureTask
	ft.Callback = Callback
	ft.SubName = SubName
	ft.GameTimeMs = TimeToFireMs + gs.GameTimeMs
	ft.Value = Value
	For i = FutureTasks.Size - 1 To 0 Step -1
		Dim old As X2FutureTask = FutureTasks.Get(i)
		If old.GameTimeMS > ft.GameTimeMs Then
			If i = FutureTasks.Size - 1 Then
				FutureTasks.Add(ft)
			Else
				FutureTasks.InsertAt(i + 1, ft)
			End If
			Return
		End If
	Next
	FutureTasks.InsertAt(0, ft)
End Sub


'Converts a bitmap to a BC.
'Scale = bitmap scale.
Public Sub BitmapToBC(bmp As B4XBitmap, Scale As Float) As BitmapCreator
	Dim b2 As BitmapCreator
	b2.Initialize(bmp.Width / Scale, bmp.Height / Scale)
	b2.CopyPixelsFromBitmap(bmp)
	Return b2
End Sub

'Converts a bitmap to a CompressedBC. 
Public Sub BitmapToCompressedBC (bmp As B4XBitmap, Scale As Float) As CompressedBC
	Dim bc As BitmapCreator = BitmapToBC(bmp, Scale)
	Return bc.ExtractCompressedBC(bc.TargetRect, GraphicCache.CBCCache)
End Sub

Private Sub CopyNewX2ScaledBitmap  (src As X2ScaledBitmap, bc As BitmapCreator) As X2ScaledBitmap
	Dim sb As X2ScaledBitmap
	sb.Scale = src.Scale
	sb.Bmp = bc.Bitmap
	Return sb
End Sub

'Horizontally flips a ScaledBitmap.
Public Sub FlipHorizontal (src As X2ScaledBitmap) As X2ScaledBitmap
	Dim bc As BitmapCreator = BitmapToBC(src.Bmp, 1)
	Dim c As Float = (bc.mWidth - 1) / 2
	Dim argbLeft, argbRight As ARGBColor
	For x = 0 To c - 0.5
		For y = 0 To bc.mHeight - 1
			bc.GetARGB(x, y, argbLeft)
			bc.GetARGB(bc.mWidth -1 - x, y, argbRight)
			bc.SetARGB(x, y,argbRight)
			bc.SetARGB(bc.mWidth -1 - x, y, argbLeft)
		Next
	Next
	Return CopyNewX2ScaledBitmap(src, bc)
End Sub

'Vertically flips a ScaledBitmap.
Public Sub FlipVertical (src As X2ScaledBitmap) As X2ScaledBitmap
	Dim bc As BitmapCreator = BitmapToBC(src.Bmp, 1)
	Dim c As Float = (bc.mHeight - 1) / 2
	Dim argbTop, argbDown As ARGBColor
	For x = 0 To bc.mWidth - 1
		For y = 0 To c
			bc.GetARGB(x, y, argbTop)
			bc.GetARGB(x, bc.mHeight - 1 - y, argbDown)
			bc.SetARGB(x, y,argbDown)
			bc.SetARGB(x, bc.mHeight - 1 - y, argbTop)
		Next
	Next
	Return CopyNewX2ScaledBitmap(src, bc)
End Sub

'Helper method to get the "other" fixture in a collision.
Public Sub GetOtherFixture (ThisBody As B2Body, Contact As B2Contact) As B2Fixture
	If Contact.FixtureA.Body = ThisBody Then Return Contact.FixtureB Else Return Contact.FixtureA
End Sub

'Helper method. Randomly returns a float number.
Public Sub RndFloat (FromValue As Float, ToValue As Float) As Float
	Return Rnd(FromValue * 100000, ToValue * 100000) / 100000
End Sub

'Helper method to create vectors.
Public Sub CreateVec2 (x As Float, y As Float) As B2Vec2
	Return World.CreateVec2 (x, y)
End Sub

'Creates a DrawTask from a CompressedBC.
Public Sub CreateDrawTaskFromCompressedBC (CBC As CompressedBC, ScreenPosition As B2Vec2) As DrawTask
	Dim dt As DrawTask = MainBC.CreateDrawTask(CBC, CBC.TargetRect, ScreenPosition.X - CBC.TargetRect.Width / 2, _
		ScreenPosition.Y - CBC.TargetRect.Height / 2, False)
	dt.IsCompressedSource = True
	Return dt
End Sub

'In B4A B4XCanvas.CreateBitmap returns the same internal mutable bitmap every time. This can be problematic if you reuse the Canvas as it will modify the bitmap.
Public Sub CreateImmutableBitmap (CVS As B4XCanvas) As B4XBitmap
	Dim bmp As B4XBitmap = CVS.CreateBitmap
	#if B4A
	Dim b As Bitmap
	b.Initialize3(bmp)
	bmp = b
	#End If
	Return bmp
End Sub

Private Sub ConvertMillisecondsToString(t As Long) As String
   Dim minutes As Int = t / DateTime.TicksPerMinute
   Dim seconds As Int = (t Mod DateTime.TicksPerMinute) / DateTime.TicksPerSecond
   Return $"$1.0{minutes}:$2.0{seconds}"$
End Sub